use anyhow::Result;
use argon2::{
    password_hash::{rand_core::OsRng, SaltString},
    Argon2, PasswordHash, PasswordHasher, PasswordVerifier,
};

pub fn random() -> String {
    SaltString::generate(&mut OsRng).to_string()
}

pub fn hash(plaintext: &str) -> Result<String> {
    let salt = SaltString::generate(&mut OsRng);
    Ok(Argon2::default()
        .hash_password(plaintext.as_bytes(), &salt)?
        .to_string())
}

pub fn verify(plaintext: &str, hash: &str) -> Result<bool> {
    let parsed_hash = PasswordHash::new(hash)?;
    Ok(Argon2::default()
        .verify_password(plaintext.as_bytes(), &parsed_hash)
        .is_ok())
}

#[cfg(test)]
mod tests {
    use std::collections::HashSet;

    #[test]
    fn test_hash_and_verify() {
        // Create random password
        let random = super::random();

        // Hash password
        let hash = super::hash(&random).unwrap();

        // Verify should be true
        assert!(super::verify(&random, &hash).unwrap());
    }

    #[test]
    fn test_random_requirements() {
        // Test that a large number of random strings are unique
        const NUM_STRINGS: usize = 10000;

        let mut strings = HashSet::new();

        for _ in 0..NUM_STRINGS {
            let random_string = super::random();

            // Strings should also be long enough
            assert!(random_string.len() >= 20);

            assert!(
                !strings.contains(&random_string),
                "Duplicate string found: {}",
                random_string
            );
            strings.insert(random_string);
        }
    }
}
