use anyhow::anyhow;
use async_trait::async_trait;
use axum::http::StatusCode;
use chrono::{Duration, NaiveDateTime};
use cookie::Cookie;
use serde::{Deserialize, Serialize};
use sqlx::FromRow;
use uuid::Uuid;

use crate::{content::Error, http::error::ApiError};

use super::user::User;

pub const USER_NOT_FOUND: ApiError<'static> = ApiError::Client {
    status: StatusCode::UNAUTHORIZED,
    code: "user-not-found",
    message: "The logged-in user was not found",
};

#[derive(Deserialize, Serialize, Clone, FromRow)]
pub struct Session {
    /// Role ID
    pub id: Uuid,

    /// User ID
    pub actor: Uuid,

    /// Secret
    pub secret: String,

    /// Times
    pub created_at: NaiveDateTime,
    pub expires_at: NaiveDateTime,
}

#[async_trait]
pub trait SessionRepository {
    async fn create_session(&self, user_id: Uuid, duration: Duration) -> Result<Session, Error>;

    async fn find_session(&self, session_id: Uuid) -> Result<Option<(Session, User)>, Error>;

    async fn refresh_session(&self, session: Session, duration: Duration)
        -> Result<Session, Error>;

    async fn destroy_session(&self, session_id: Uuid) -> Result<(), Error>;

    async fn prune_dead_sessions(&self) -> Result<u64, Error>;
}

impl Session {
    pub fn token(&self) -> String {
        format!("{}:{}", self.id.as_u128(), self.secret)
    }

    pub fn parse_token(token: &str) -> Result<(Uuid, String), Error> {
        let (uuid_str, token_str) = token
            .split_once(':')
            .ok_or_else(|| Error::Internal(anyhow!("malformed token")))?;
        Ok((
            Uuid::from_u128(
                uuid_str
                    .parse::<u128>()
                    .map_err(|e| Error::Internal(e.into()))?,
            ),
            token_str.to_string(),
        ))
    }

    pub fn cookie(&self, domain: &str, secure: bool) -> String {
        Cookie::build("session", self.token())
            .domain(domain)
            .secure(secure)
            .http_only(!secure)
            .path("/")
            .expires(
                cookie::time::OffsetDateTime::from_unix_timestamp(self.expires_at.timestamp())
                    .unwrap(),
            )
            .finish()
            .to_string()
    }

    pub fn cookie_for_delete(domain: &str, secure: bool) -> String {
        Cookie::build("session", "")
            .domain(domain)
            .secure(secure)
            .http_only(!secure)
            .path("/")
            .max_age(cookie::time::Duration::seconds(0))
            .finish()
            .to_string()
    }
}
