use async_trait::async_trait;
use chrono::NaiveDateTime;
use serde::{Deserialize, Serialize};
use sqlx::FromRow;
use uuid::Uuid;

use crate::content::Error;

#[derive(Deserialize, Serialize, Clone, FromRow)]
pub struct User {
    /// User internal ID
    pub id: Uuid,

    /// User name (unique per instance, shows up in URLs)
    pub name: String,

    /// User email (for validation/login)
    pub email: Option<String>,

    /// Hashed password
    pub password: Option<String>,

    /// User's chosen displayed name
    pub display_name: Option<String>,

    /// Biography / User description
    pub bio: Option<String>,

    /// User roles (as role IDs)
    pub roles: Vec<Uuid>,

    /// Times
    pub created_at: NaiveDateTime,
    pub modified_at: Option<NaiveDateTime>,
    pub deleted_at: Option<NaiveDateTime>,
}

impl Default for User {
    fn default() -> Self {
        Self {
            id: Uuid::nil(),
            name: Default::default(),
            email: Default::default(),
            password: Default::default(),
            display_name: Default::default(),
            bio: Default::default(),
            roles: Default::default(),
            created_at: Default::default(),
            modified_at: Default::default(),
            deleted_at: Default::default(),
        }
    }
}

#[async_trait]
pub trait UserRepository {
    async fn create_user(
        &self,
        username: &str,
        password: &str,
        roles: &Vec<Uuid>,
    ) -> Result<User, Error>;

    async fn find_user(&self, name: &str) -> Result<Option<User>, Error>;

    async fn has_no_users(&self) -> Result<bool, Error>;
}

#[derive(Deserialize, Serialize, FromRow)]
pub struct Role {
    /// Role ID
    pub id: Uuid,

    /// Role scopes (permissions)
    pub scopes: Vec<String>,
}
