use async_trait::async_trait;
use chrono::{Duration, Utc};
use uuid::Uuid;

use crate::{
    auth::{
        hash::random,
        session::{Session, SessionRepository},
        user::User,
    },
    content::Error,
};

use super::Database;

#[async_trait]
impl SessionRepository for Database {
    async fn create_session(&self, user_id: Uuid, duration: Duration) -> Result<Session, Error> {
        let now = Utc::now().naive_utc();
        let expires = now + duration;
        let secret = random();
        let result = sqlx::query!(
		"INSERT INTO sessions (id, actor, secret, created_at, expires_at) VALUES ($1, $2, $3, $4, $5) RETURNING id",
		Uuid::now_v7(),
		user_id,
		secret,
		now,
		expires
	    )
	    .fetch_one(&self.pool)
	    .await?;
        Ok(Session {
            id: result.id,
            actor: user_id,
            secret,
            created_at: now,
            expires_at: now + duration,
        })
    }

    async fn find_session(&self, session_id: Uuid) -> Result<Option<(Session, User)>, Error> {
        let record = sqlx::query!(
            "SELECT
			sessions.id AS session_id,
			sessions.actor AS session_actor,
			sessions.secret,
			sessions.created_at AS session_created_at,
			sessions.expires_at,
			users.id AS user_id,
			users.name,
			users.email,
			users.display_name,
			users.bio,
			users.roles,
			users.created_at AS user_created_at,
			users.modified_at,
			users.deleted_at
		    FROM
			sessions
		    JOIN
			users ON sessions.actor = users.id
		    WHERE
			sessions.id = $1",
            session_id
        )
        .fetch_optional(&self.pool)
        .await?;

        Ok(record.map(|record| {
            (
                Session {
                    id: record.session_id,
                    actor: record.session_actor,
                    secret: record.secret,
                    created_at: record.session_created_at,
                    expires_at: record.expires_at,
                },
                User {
                    id: record.user_id,
                    name: record.name,
                    email: record.email,
                    password: None,
                    display_name: record.display_name,
                    bio: record.bio,
                    roles: record.roles,
                    created_at: record.user_created_at,
                    modified_at: record.modified_at,
                    deleted_at: record.deleted_at,
                },
            )
        }))
    }

    async fn refresh_session(
        &self,
        session: Session,
        duration: Duration,
    ) -> Result<Session, Error> {
        let expires_at = (Utc::now() + duration).naive_utc();

        sqlx::query!(
            "UPDATE sessions SET expires_at = $1 WHERE id = $2 RETURNING id",
            expires_at,
            session.id
        )
        .fetch_one(&self.pool)
        .await?;

        Ok(Session {
            expires_at,
            ..session
        })
    }

    async fn destroy_session(&self, session_id: Uuid) -> Result<(), Error> {
        sqlx::query!("DELETE FROM sessions WHERE id = $1", session_id)
            .execute(&self.pool)
            .await?;
        Ok(())
    }

    async fn prune_dead_sessions(&self) -> Result<u64, Error> {
        let now = Utc::now().naive_utc();
        let result = sqlx::query!("DELETE FROM sessions WHERE expires_at < $1", now)
            .execute(&self.pool)
            .await?;
        Ok(result.rows_affected())
    }
}
