use async_trait::async_trait;
use uuid::Uuid;

use crate::{
    auth::hash::hash,
    auth::user::{User, UserRepository},
    content::Error,
};

use super::Database;

#[async_trait]
impl UserRepository for Database {
    async fn create_user(
        &self,
        username: &str,
        password: &str,
        roles: &Vec<Uuid>,
    ) -> Result<User, Error> {
        let result = sqlx::query!(
            r#"INSERT INTO users ( id, name, password, roles )
		    VALUES ( $1,$2,$3,$4 ) RETURNING id, created_at"#,
            Uuid::now_v7(),
            username,
            hash(&password).map_err(|e| Error::Internal(e))?,
            roles,
        )
        .fetch_one(&self.pool)
        .await?;
        Ok(User {
            id: result.id,
            name: username.to_owned(),
            roles: roles.to_owned(),
            created_at: result.created_at,
            ..Default::default()
        })
    }

    async fn find_user(&self, name: &str) -> Result<Option<User>, Error> {
        Ok(sqlx::query_as("SELECT * FROM users WHERE name = $1")
            .bind(name)
            .fetch_optional(&self.pool)
            .await?)
    }

    async fn has_no_users(&self) -> Result<bool, Error> {
        // Check if the user table is completely empty
        let empty = sqlx::query!(
            "SELECT CASE WHEN EXISTS(SELECT 1 FROM users) THEN false ELSE true END AS empty;"
        )
        .map(|row| row.empty.unwrap_or(true))
        .fetch_one(&self.pool)
        .await?;

        Ok(empty)
    }
}
