use axum::{
    extract::rejection::JsonRejection,
    http::StatusCode,
    response::{IntoResponse, Response},
    Json,
};
use serde_json::json;
use thiserror::Error;

use crate::content;

// Generic errors
const ERR_NOT_FOUND: ApiError<'static> = ApiError::Client {
    status: StatusCode::NOT_FOUND,
    code: "not-found",
    message: "resource not found",
};
const ERR_NOT_AVAILABLE: ApiError<'static> = ApiError::Client {
    status: StatusCode::CONFLICT,
    code: "id-not-available",
    message: "the chosen identifier is not available",
};
const ERR_UNAUTHORIZED: ApiError<'static> = ApiError::Client {
    status: StatusCode::FORBIDDEN,
    code: "access-denied",
    message: "you are not authorized to perform this action",
};

#[derive(Error, Debug)]
pub enum ApiError<'a> {
    #[error("client error: <{code}> {message}")]
    Client {
        status: StatusCode,
        code: &'a str,
        message: &'a str,
    },

    #[error("unexpected internal error: {0}")]
    Internal(#[from] anyhow::Error),

    #[error("database returned error: {0}")]
    Database(#[from] sqlx::Error),

    #[error("incoming JSON format error: {0}")]
    JSONFormat(#[from] JsonRejection),
}

impl IntoResponse for ApiError<'_> {
    fn into_response(self) -> Response {
        match self {
            ApiError::Internal(err) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                Json(json!({"code":"server-error", "message": err.to_string()})),
            )
                .into_response(),
            ApiError::Database(err) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                Json(json!({"code":"server-error", "message": err.to_string()})),
            )
                .into_response(),
            ApiError::Client {
                status,
                code,
                message,
            } => (status, Json(json!({"code":code, "message": message}))).into_response(),
            ApiError::JSONFormat(rejection) => {
                let status = match rejection {
                    JsonRejection::JsonDataError(_) => StatusCode::UNPROCESSABLE_ENTITY,
                    JsonRejection::JsonSyntaxError(_) => StatusCode::BAD_REQUEST,
                    JsonRejection::MissingJsonContentType(_) => StatusCode::UNSUPPORTED_MEDIA_TYPE,
                    _ => StatusCode::INTERNAL_SERVER_ERROR,
                };
                (
                    status,
                    Json(json!({"code":"invalid-body", "message": rejection.body_text()})),
                )
                    .into_response()
            }
        }
    }
}

impl From<content::Error> for ApiError<'_> {
    fn from(err: content::Error) -> Self {
        match err {
            content::Error::NotFound => ERR_NOT_FOUND,
            content::Error::IdentifierNotAvailable => ERR_NOT_AVAILABLE,
            content::Error::AccessDenied => ERR_UNAUTHORIZED,
            content::Error::QueryFailed(err) => err.into(),
            content::Error::Internal(err) => err.into(),
        }
    }
}
