use axum::http::StatusCode;
use axum::response::IntoResponse;
use axum::routing::post;
use axum::Json;
use axum::Router;
use serde::Serialize;
use serde_json::json;
use std::sync::Arc;
use utoipa::ToSchema;

use crate::auth::user::UserRepository;
use crate::database::Database;
use crate::{
    auth::hash::random, builtins::ROLE_SUPERADMIN, http::error::ApiError, state::AppState,
};

#[derive(Serialize, ToSchema)]
pub(super) struct BootstrapInfo {
    /// Admin username
    pub username: String,

    /// Admin password
    pub password: String,
}

/// Bootstrap the instance by creating a admin user
#[utoipa::path(
    post,
    path = "/admin/bootstrap",
    responses(
        (status = 200, description = "List of collections fetched", body = [CollectionData])
    )
)]
pub(super) async fn bootstrap<Repo: UserRepository>(repository: Repo) -> impl IntoResponse {
    // Only allow this request if the user table is completely empty!
    if !repository.has_no_users().await? {
        return Err(ApiError::Client {
            status: StatusCode::BAD_REQUEST,
            code: "already-setup",
            message: "The instance was already bootstrapped",
        });
    }

    let username = "admin".to_string();
    let password = random();

    repository
        .create_user(&username, &password, &[ROLE_SUPERADMIN].to_vec())
        .await?;

    Ok(Json(json!(BootstrapInfo { username, password })))
}

pub(super) fn router() -> Router<Arc<AppState>> {
    Router::new().route("/bootstrap", post(bootstrap::<Database>))
}
