use axum::{
    extract::Path,
    response::IntoResponse,
    routing::{get, post},
    Json, Router,
};
use serde_json::json;
use std::sync::Arc;

use crate::{
    content::{
        post::{CreatePostData, PostRepository, UpdatePostData},
        site::SiteRepository,
        Error,
    },
    database::Database,
    http::{
        error::ApiError,
        json::JsonBody,
        session::{OptionalUser, RequireUser},
    },
    state::AppState,
};

/// Retrieve post by site and post slug
#[utoipa::path(
    get,
    path = "/posts/{site}/{slug}",
    params(
        ("site" = String, Path, description = "Site slug"),
        ("slug" = String, Path, description = "Post slug")
    ),
    responses(
        (status = 200, description = "Post found", body = PostWithAuthor),
        (status = 404, description = "Post not found")
    )
)]
pub(super) async fn get_post<Repo: PostRepository>(
    repository: Repo,
    OptionalUser(user): OptionalUser,
    Path((site, slug)): Path<(String, String)>,
) -> Result<impl IntoResponse, ApiError<'static>> {
    let post = repository.get_post_from_url(&site, &slug).await?;

    // If post has been published, show to everyone, otherwise only show to owner/author
    if post.published {
        return Ok(Json(post));
    }

    if let Some(user) = user {
        // TODO This will need to be changed if we want more users managing the same site
        // Also, we should have a better way to match the user!! (mapping func? prolly)
        if user.name == post.author_username {
            return Ok(Json(post));
        }
    }

    // 404'd!!
    Err(Error::NotFound.into())
}

/// Create a new post
#[utoipa::path(
    post,
    path = "/posts/{site}",
    params(
        ("site" = String, Path, description = "Site slug"),
    ),
    request_body = CreatePostData,
    responses(
        (status = 200, description = "Post created successfully"),
        (status = 403, description = "You don't have the necessary permissions to create posts on this site"),
        (status = 409, description = "The chosen post identifier/slug is not available")
    )
)]
pub(super) async fn create_post<Repo: PostRepository + SiteRepository>(
    repository: Repo,
    Path(site): Path<String>,
    RequireUser(user): RequireUser,
    JsonBody(data): JsonBody<CreatePostData>,
) -> Result<impl IntoResponse, ApiError<'static>> {
    // Resolve site
    let (site_id, owner) = repository.get_site_id_and_owner(&site).await?;

    if user.id != owner {
        return Err(Error::AccessDenied.into());
    }

    repository.create_post(&user.id, &site_id, data).await?;
    Ok(Json(json!({ "ok": true })))
}

/// Update an existing post
#[utoipa::path(
    put,
    path = "/posts/{site}/{slug}",
    params(
        ("site" = String, Path, description = "Site slug"),
        ("slug" = String, Path, description = "Post slug")
    ),
    request_body = UpdatePostData,
    responses(
        (status = 200, description = "Post updated successfully"),
        (status = 403, description = "You don't have the necessary permissions to update posts on this site"),
        (status = 404, description = "Post not found")
    )
)]
pub(super) async fn update_post<Repo: PostRepository + SiteRepository>(
    repository: Repo,
    Path((site, slug)): Path<(String, String)>,
    RequireUser(user): RequireUser,
    JsonBody(data): JsonBody<UpdatePostData>,
) -> Result<impl IntoResponse, ApiError<'static>> {
    // Resolve site
    let (site_id, owner) = repository.get_site_id_and_owner(&site).await?;

    if user.id != owner {
        return Err(Error::AccessDenied.into());
    }

    repository
        .update_post(&user.id, &site_id, &slug, data)
        .await?;
    Ok(Json(json!({ "ok": true })))
}

/// Delete an existing post
#[utoipa::path(
    delete,
    path = "/posts/{site}/{slug}",
    params(
        ("site" = String, Path, description = "Site slug"),
        ("slug" = String, Path, description = "Post slug")
    ),
    responses(
        (status = 200, description = "Post deleted successfully"),
        (status = 403, description = "You don't have the necessary permissions to delete posts on this site"),
        (status = 404, description = "Post not found")
    )
)]
pub(super) async fn delete_post<Repo: PostRepository + SiteRepository>(
    repository: Repo,
    Path((site, slug)): Path<(String, String)>,
    RequireUser(user): RequireUser,
) -> Result<impl IntoResponse, ApiError<'static>> {
    // Resolve site
    let (site_id, owner) = repository.get_site_id_and_owner(&site).await?;

    if user.id != owner {
        return Err(Error::AccessDenied.into());
    }

    repository
        .delete_post(&user.id, &site_id, &slug, true)
        .await?;
    Ok(Json(json!({ "ok": true })))
}

pub(super) fn router() -> Router<Arc<AppState>> {
    Router::new()
        .route("/:site", post(create_post::<Database>))
        .route(
            "/:site/:slug",
            get(get_post::<Database>)
                .put(update_post::<Database>)
                .delete(delete_post::<Database>),
        )
}
